library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity sdcard_controller is
	port (
		clock		: in  std_logic;
		reset    : in  std_logic;
		sck		: out std_logic 		:= '0';
		cmd      : inout  std_logic   := '1';
		dat0     : inout  std_logic   := '1';
		dat1     : inout  std_logic  	:= '1';
		dat2     : inout  std_logic   := '1';
		dat3     : inout  std_logic 	:= '1';
		ready		: out std_logic := '0';
		read_addr : in std_logic_vector(31 downto 0);
		read_enable : in std_logic;
		data		: out std_logic_vector(7 downto 0) := (others => '0');
		data_ready : out std_logic := '0';
		debug_await_data_cmd : out std_logic := '0';
		debug_await_data : out std_logic := '0';
		debug_card_error : out std_logic := '0'
	);
end entity sdcard_controller;

architecture rtl of sdcard_controller is
-- 0 = initialization speed 400KHz; 1 = High speed	25MHz
	constant CLOCK_DIV_INIT : natural range 0 to 126 := 125; -- we assume clock is 100MHz -> ~400KHz
	constant CLOCK_DIV_HS : natural range 0 to 2 := 2; -- we assume clock is 100MHz -> 25MHz -> turns out with the default main clock of 50 MHz it only works with a ~ 8.3333333MHz card clock
	signal clock_div : natural range 0 to CLOCK_DIV_INIT := 0;
	signal clock_div_counter :natural range 0 to CLOCK_DIV_INIT := 0;
	signal card_clock : std_logic := '0';
	signal clock_enable : std_logic := '0';
	signal clock_state :std_logic_vector(1 downto 0) := (others => '0');
	
	type tcontrollerstate is (no_card, card_init, card_idle, card_get_blockread, card_await_data, card_switch_clock, card_wait8, card_send_asc, card_send_asc_w_rca, card_get_asc, card_send_cmd, card_send_cmd_w_rsp, card_prep_resp, card_await_response, card_get_response, card_get_response_long,  card_error);
	signal controllerstate : tcontrollerstate := no_card;
	signal nextcontrollerstate : tcontrollerstate := no_card;
	signal card_cmd_asc : std_logic := '0';
	
	signal waitcounter : natural range 0 to 128 := 0;
	
	component crc7 is
		port(
			clk : in std_logic;
			reset : in std_logic;
			enable : in std_logic;
			input : in std_logic;
			crc : out std_logic_vector(6 downto 0)
		);
	end component crc7;
	signal crc_enable : std_logic := '0';
	signal crc_reset : std_logic := '0';
	signal crc_output : std_logic_vector(6 downto 0) := (others => '0');
	
	constant CMD0 : std_logic_vector(39 downto 0)   := "01"&"000000"&"00000000"&"00000000"&"00000000"&"00000000";
	constant CMD8 : std_logic_vector(39 downto 0)   := "01"&"001000"&"00000000"&"00000000"&"00000001"&"10101010";
	constant RESP8 : std_logic_vector(47 downto 0)  := "00"&"001000"&"00000000"&"00000000"&"00000001"&"10101010"&"0001001"&"1";
	constant CMD55 : std_logic_vector(39 downto 0)  := "01"&"110111"&"00000000"&"00000000"&"00000000"&"00000000";
	constant RESP55 : std_logic_vector(7 downto 0)  := "00"&"110111";
	constant CMD41 : std_logic_vector(39 downto 0)  := "01"&"101001"&"01000000"&"00010000"&"00000000"&"00000000";
	constant CMD2 : std_logic_vector(39 downto 0)   := "01"&"000010"&"00000000"&"00000000"&"00000000"&"00000000";
	constant CMD3 : std_logic_vector(39 downto 0)   := "01"&"000011"&"00000000"&"00000000"&"00000000"&"00000000";
	constant CMD9 : std_logic_vector(7 downto 0)    := "01"&"001001";
	constant CMD7 : std_logic_vector(7 downto 0)    := "01"&"000111";
	constant CMD42 : std_logic_vector(7 downto 0)   := "01"&"101010";
	constant CMD6 : std_logic_vector(7 downto 0)    := "01"&"000110";
	constant CMD16 : std_logic_vector(39 downto 0)  := "01"&"010000"&"00000000"&"00000000"&"00000010"&"00000000";
	constant CMD17 : std_logic_vector(7 downto 0)   := "01"&"010001";
	
	signal command : std_logic_vector(47 downto 0) := (others => '0');
	signal last_command : std_logic_vector(5 downto 0) := (others => '0');
	signal command_bit_counter : natural range 0 to 47 := 0;
	signal response : std_logic_vector(47 downto 0) := (others => '0');
	signal response_long : std_logic_vector(135 downto 0) := (others => '0');
	signal response_bit_counter : natural range 0 to 135 := 0;
	signal response_timeout_counter : natural range 0 to 255 := 0;
	
	signal is_high_capacity : std_logic := '0';
	signal resp_type : std_logic := '0';
	signal cid :std_logic_vector(127 downto 0) := (others => '0');
	signal rca :std_logic_vector(15 downto 0) := (others => '0');
	signal csd :std_logic_vector(127 downto 0) := (others => '0');
	
	type tinitstate is (init_rst_sic, init_get_sic, init_asc, init_get_asc, init_ini, init_check_ready, init_get_cid, init_get_rca, init_get_csd, init_switch_clock, init_transfermode, init_get_transfermode, init_disable_pullup, init_buswidth, init_setbuswidth, init_setblocklength, init_complete);
	signal initstate : tinitstate := init_rst_sic;
	
	signal data_counter : natural range 0 to 1032 := 0;
	signal data_byte : std_logic_vector(7 downto 0) := (others => '0');
	signal data_hl : std_logic := '0';
	signal data_timeout_counter : natural range 0 to 65535 := 0;
	
	type tdatastate is (data_idle, data_await, data_timeout, data_readh, data_readl, data_avail, data_finished);
	signal datastate : tdatastate := data_idle;
	
begin
-- sd card clock
	sck <= card_clock;
									 
-- clock_proc	
	clock_proc: process
	begin
		wait until rising_edge(clock);
			if (reset = '1') then
				clock_div_counter <= 0;
				card_clock <= '0';
				clock_state <= (others => '0');
			else
				if (clock_enable = '1') then
					if (clock_div_counter < clock_div) then
						clock_div_counter <= clock_div_counter + 1;
					else
						clock_div_counter <= 0;
						card_clock <= not card_clock;
					end if;
					clock_state <= clock_state(0) & card_clock;
				else
					clock_state <= (others => '0');
					card_clock <= '0';
					clock_div_counter <= 0;
				end if;
			end if;
	end process;
	
	cmd_crc7 : crc7 port map(
		clk => clock,
		reset => crc_reset,
		enable => crc_enable,
		input => cmd,
		crc => crc_output
	);	
	
-- data transfer fsm
	data_fsm : process
	begin
		wait until rising_edge(clock);
		if (reset = '1') then
			data_ready <= '0';
			data_timeout_counter <= 0;
			data_counter <= 0;
			data <= (others => '0');
			dat0 <= '1';
			dat1 <= '1';
			dat2 <= '1';
			dat3 <= '1';
			datastate <= data_idle;
			debug_await_data <= '0';
		else
			data_ready <= '0';
			case datastate is
				when data_idle =>
					debug_await_data <= '0';
					data_timeout_counter <= 0;
					dat0 <= '1';
					dat1 <= '1';
					dat2 <= '1';
					dat3 <= '1';
					if (controllerstate = card_await_data) then
						dat0 <= 'Z';
						dat1 <= 'Z';
						dat2 <= 'Z';
						dat3 <= 'Z';
						datastate <= data_await;
					end if;
					
				when data_await =>
					debug_await_data <= '1';
					if (clock_state = "01") then
						if ((dat0 = '0') and (dat1 = '0') and (dat2 = '0') and (dat3 = '0')) then
							data_counter <= 0;
							data_byte <= (others => '0');
							datastate <= data_readh;
							debug_await_data <= '0';
						else
							if (data_timeout_counter < 65535) then
								data_timeout_counter <= data_timeout_counter +1;
							else
								datastate <= data_timeout;
							end if;
						end if;
					end if;
					
				when data_timeout =>
					if (controllerstate = no_card) then
						datastate <= data_idle;
					end if;
				
				when data_readh =>
					if (clock_state = "01") then
						data_byte(7 downto 4) <= dat3 & dat2 & dat1 & dat0;
						datastate <= data_readl;
					end if;
					
				when data_readl =>
					if (clock_state = "01") then
						data_byte(3 downto 0) <= dat3 & dat2 & dat1 & dat0;
						datastate <= data_avail;
					end if;
					
				when data_avail =>
					if (data_counter < 512) then
						data <= data_byte;
						data_ready <= '1';
					end if;
					if (data_counter < 513) then
						data_counter <= data_counter + 1;
						datastate <= data_readh;
					else
						datastate <= data_finished;
					end if;
				
				when data_finished =>
					datastate <= data_idle;
				
				when others =>
					datastate <= data_idle;
				
			end case;
		end if;
	end process;
	
-- controller FSM	
	controller_fsm :process
	begin
		wait until rising_edge(clock);
		if (reset = '1') then
			controllerstate <= no_card;
			nextcontrollerstate <= no_card;
			initstate <= init_rst_sic;
			card_cmd_asc <= '0';
			resp_type <= '0';
			clock_enable <= '0';
			clock_div <= CLOCK_DIV_INIT;
			waitcounter <= 0;
			cmd <= '1';
			command(47 downto 8) <= CMD0;
			cid <= (others => '0');
			rca <= (others => '0');
			csd <= (others => '0');
			response <= (others => '0');
			response_long <= (others => '0');
			response_bit_counter <= 0;
			is_high_capacity <= '0';
			crc_reset <= '1';
			ready <= '0';
			debug_card_error <= '0';
			debug_await_data_cmd <= '0';
		else
			
			crc_enable <= '0';
			clock_enable <= '1';
			
			case controllerstate is
				when no_card =>
					debug_card_error <= '0';
					debug_await_data_cmd <= '0';
					if (clock_state = "10") then
						if (waitcounter < 80) then
							waitcounter <= waitcounter + 1;
						else
							crc_reset <= '0';
							command(47 downto 8) <= CMD0; -- send soft reset command (CMD0, no response)
							command_bit_counter <= 0;
							card_cmd_asc <= '0';
							resp_type <= '0';
							initstate <= init_rst_sic;
							nextcontrollerstate <= card_init;
							controllerstate <= card_send_cmd;
						end if;					
					end if;
					
				when card_init =>
					case initstate is			
						when init_rst_sic =>
							command(47 downto 8) <= CMD8; -- send voltage and check pattern
							command_bit_counter <= 0;
							card_cmd_asc <= '0';
							resp_type <= '0';
							initstate <= init_get_sic;
							controllerstate <= card_send_cmd_w_rsp;
							nextcontrollerstate <= card_init;
							
						when init_get_sic =>
							if (response = RESP8) then
								initstate <= init_asc;
							else
								controllerstate <= card_error;
							end if;
							
						when init_asc =>
							initstate <= init_ini;
							controllerstate <= card_send_asc;  -- send next command is application specific
							nextcontrollerstate <= card_init;
						
						when init_ini =>
							command(47 downto 8) <= CMD41; -- next command is activate card ini process
							command_bit_counter <= 0;
							card_cmd_asc <= '0';
							resp_type <= '0';
							initstate <= init_check_ready;
							controllerstate <= card_send_cmd_w_rsp;
							nextcontrollerstate <= card_init;				
							
						when init_check_ready =>
							if ((response(47 downto 39) = "001111111") and (response(7 downto 0) = "11111111")) then
								is_high_capacity <= response(38);
								if (response(28) = '1') then
									command(47 downto 8) <= CMD2; -- get long cid csd response
									command_bit_counter <= 0;
									card_cmd_asc <= '0';
									resp_type <= '1';
									controllerstate <= card_send_cmd_w_rsp;
									initstate <= init_get_cid;
									nextcontrollerstate <= card_init;
								else
									controllerstate <= card_error;
								end if;
							else
								initstate <= init_asc;
								waitcounter <= 0;
								controllerstate <= card_wait8;
								nextcontrollerstate <= card_init;
							end if;
						
						when init_get_cid =>
							if (response_long(135 downto 128) = "00111111") then
								cid <= response_long(127 downto 0);
								command(47 downto 8) <= CMD3; -- get RCA
								command_bit_counter <= 0;
								card_cmd_asc <= '0';
								resp_type <= '0';
								controllerstate <= card_send_cmd_w_rsp;
								initstate <= init_get_rca;
								nextcontrollerstate <= card_init;
							else
								controllerstate <= card_error;
							end if;
								
						when init_get_rca =>
							if ((response(47 downto 40) = "00000011") and (response(23 downto 21) = "000")) then
								rca <= response(39 downto 24);
								command(47 downto 40) <= CMD9; -- get CSD
								command(39 downto 8) <= response(39 downto 24) & "0000000000000000"; -- RCA
								command_bit_counter <= 0;
								card_cmd_asc <= '0';
								resp_type <= '1';
								controllerstate <= card_send_cmd_w_rsp;
								initstate <= init_get_csd;
								nextcontrollerstate <= card_init;
							else
								controllerstate <= card_error;
							end if;
						
						when init_get_csd =>
							if (response_long(135 downto 128) = "00111111") then
								csd <= response_long(127 downto 0);
								if (response_long(127 downto 126) /= "01") then -- we only want csd structure version 2.0
									controllerstate <= card_error;
								else 
									waitcounter <= 0;
									controllerstate <= card_wait8;
									initstate <= init_switch_clock;
									nextcontrollerstate <= card_init;
								end if;							
							else
								controllerstate <= card_error;
							end if;
						
						when init_switch_clock =>
							controllerstate <= card_switch_clock;
							initstate <= init_transfermode;
							nextcontrollerstate <= card_init;
						
						when init_transfermode =>
							command(47 downto 40) <= CMD7; -- set cardto transfermode
							command(39 downto 8) <= rca & "0000000000000000"; -- RCA
							command_bit_counter <= 0;
							card_cmd_asc <= '0';
							resp_type <= '0';
							controllerstate <= card_send_cmd_w_rsp;
							initstate <= init_get_transfermode;
							nextcontrollerstate <= card_init;
						
						when init_get_transfermode =>
							if ((response(47 downto 40) = "00000111") and (response(27) = '0')) then
								initstate <= init_disable_pullup;
								controllerstate <= card_send_asc_w_rca;	-- asc with rca set
								nextcontrollerstate <= card_init;
							else
								controllerstate <= card_error;
							end if;
						
						when init_disable_pullup =>
							command(47 downto 40) <= CMD42; -- disable pullup on dat3
							command(39 downto 8) <= rca & "0000000000000000"; -- RCA + pullup disabled
							command_bit_counter <= 0;
							card_cmd_asc <= '0';
							resp_type <= '0';
							controllerstate <= card_send_cmd_w_rsp;
							initstate <= init_buswidth;
							nextcontrollerstate <= card_init;
						
						when init_buswidth =>
							if ((response(47 downto 40 ) = "00101010") and (response(27) = '0')) then
								initstate <= init_setbuswidth;
								controllerstate <= card_send_asc_w_rca;	-- asc with rca set
								nextcontrollerstate <= card_init;
							else
								controllerstate <= card_error;
							end if;

						when init_setbuswidth =>
							command(47 downto 40) <= CMD6; -- set bus width
							command(39 downto 8) <= rca & "0000000000000010"; -- RCA + bus width 4 bit
							command_bit_counter <= 0;
							card_cmd_asc <= '0';
							resp_type <= '0';
							controllerstate <= card_send_cmd_w_rsp;
							initstate <= init_setblocklength;
							nextcontrollerstate <= card_init;
							
						when init_setblocklength =>
							if ((response(47 downto 40 ) = "00000110") and (response(27) = '0')) then
								command(47 downto 8) <= CMD16; -- set block length to 512 byte
								command_bit_counter <= 0;
								card_cmd_asc <= '0';
								resp_type <= '0';
								controllerstate <= card_send_cmd_w_rsp;
								initstate <= init_complete;
								nextcontrollerstate <= card_init;
							else
								controllerstate <= card_error;
							end if;
						
						when init_complete =>
							if ((response(47 downto 40 ) = "00010000") and (response(27) = '0')) then
								controllerstate <= card_idle;
							else
								controllerstate <= card_error;
							end if;
						
						when others =>
							initstate <= init_rst_sic;
							
					end case;
					
				when card_idle =>
					ready <= '1';
					if (read_enable = '1') then
						command(47 downto 40) <= CMD17; -- read single block
						if (is_high_capacity = '1') then
							command(39 downto 8) <= "000000000" & read_addr(31 downto 9); 
						else
							command(39 downto 8) <= read_addr(31 downto 9) & "000000000";
						end if;
						command_bit_counter <= 0;
						card_cmd_asc <= '0';
						resp_type <= '0';
						ready <= '0';
						controllerstate <= card_send_cmd_w_rsp;
						nextcontrollerstate <= card_get_blockread;
					end if;
				
				when card_get_blockread =>
					ready <= '0';
					if ((response(47 downto 40 ) = "00010001") and (response(27) = '0')) then
						controllerstate <= card_await_data;
					else
						controllerstate <= card_error;
					end if;
				
				when card_await_data =>
					ready <= '0';
					if (datastate = data_finished) then
						controllerstate <= card_idle;
					elsif (datastate = data_timeout) then
						controllerstate <= card_error;
					end if;
				
				when card_switch_clock =>
					if (clock_state = "10") then
						clock_div <= CLOCK_DIV_HS;
						waitcounter <= 0;
						controllerstate <= card_wait8;
					end if;			
				
				when card_wait8 =>
					if (clock_state = "10") then
						if (waitcounter < 7) then
							waitcounter <= waitcounter + 1;
						else
							controllerstate <= nextcontrollerstate;
						end if;
					end if;
					
				when card_send_asc =>
					command(47 downto 8) <= CMD55; -- next command is an Application Specific Command
					command_bit_counter <= 0;
					card_cmd_asc <= '1';
					resp_type <= '0';
					controllerstate <= card_send_cmd_w_rsp;			

				when card_send_asc_w_rca =>
					command(47 downto 40) <= CMD55(39 downto 32); -- next command is an Application Specific Command
					command(39 downto 8) <= rca & "0000000000000000";
					command_bit_counter <= 0;
					card_cmd_asc <= '1';
					resp_type <= '0';
					controllerstate <= card_send_cmd_w_rsp;			
				
				when card_get_asc =>
					if (response(47 downto 40) = RESP55) then
						controllerstate <= nextcontrollerstate;
					else
						controllerstate <= card_error;
					end if;											
				
				when card_send_cmd =>
					crc_reset <= '0';
					if (command_bit_counter = 0) then
						last_command <= command(45 downto 40);
					end if;
					if (clock_state = "10") then
						cmd <= command(47);
						command <= command(46 downto 0) & '0';
						if (command_bit_counter < 47) then
							if (command_bit_counter < 40) then
								crc_enable <= '1';
							end if;
							command_bit_counter <= command_bit_counter + 1;
						else
							crc_reset <= '1';
							waitcounter <= 0;
							controllerstate <= card_wait8;
						end if;
					end if;
					if ((clock_state = "00") and (command_bit_counter = 40)) then
						command(47 downto 40) <= crc_output & '1';
					end if;
					
				when card_send_cmd_w_rsp =>
					crc_reset <= '0';
					if (command_bit_counter = 0) then
						last_command <= command(45 downto 40);
					end if;
					if (clock_state = "10") then
						cmd <= command(47);
						command <= command(46 downto 0) & '0';
						if (command_bit_counter < 47) then
							if (command_bit_counter < 40) then
								crc_enable <= '1';
							end if;
							command_bit_counter <= command_bit_counter + 1;
						else
							crc_reset <= '1';
							controllerstate <= card_prep_resp;
						end if;
					end if;
					if ((clock_state = "00") and (command_bit_counter = 40)) then
						command(47 downto 40) <= crc_output & '1';
					end if;

				when card_prep_resp =>
					crc_reset <= '0';
					if (clock_state = "10") then
						cmd <= 'Z';
						response_timeout_counter <= 0;
						controllerstate <= card_await_response;
					end if;
				
				when card_await_response =>				
					debug_await_data_cmd <= '1';
					if (clock_state = "10") then
						if (cmd = '0') then
							response <= (others => '0');
							response_long <= (others => '0');
							response_bit_counter <= 1;
							crc_enable <= '1';
							if (resp_type = '1') then
								controllerstate <= card_get_response_long;
							else
								controllerstate <= card_get_response;
							end if;
							debug_await_data_cmd <= '0';
						else
							if (response_timeout_counter < 255) then
								response_timeout_counter <= response_timeout_counter + 1;
							else
								controllerstate <= card_error;
							end if;
						end if;
					end if;				

				when card_get_response =>
					if (clock_state = "10") then
						response <= response(46 downto 0) & cmd;
						if (response_bit_counter < 47) then
							if (response_bit_counter < 40) then
								crc_enable <= '1';
							end if;
							response_bit_counter <= response_bit_counter + 1;
						else
							if ((crc_output /= response(6 downto 0)) and (last_command /= "101001")) then
								controllerstate <= card_error;
							else
								if (card_cmd_asc = '0') then
									controllerstate <= nextcontrollerstate;
								else
									controllerstate <= card_get_asc;
								end if;
							end if;
							crc_reset <= '1';
						end if;
					end if;					

				when card_get_response_long =>
					if (clock_state = "10") then
						response_long <= response_long(134 downto 0) & cmd;
						if (response_bit_counter < 135) then
							if ((response_bit_counter > 7) and (response_bit_counter < 128)) then
								crc_enable <= '1';
							end if;						
							response_bit_counter <= response_bit_counter + 1;
						else
							if (crc_output /= response_long(6 downto 0)) then
								controllerstate <= card_error;
							else
								controllerstate <= nextcontrollerstate;
							end if;
							crc_reset <= '1';
						end if;
					end if;
					
				when card_error =>
					ready <= '0';
					debug_card_error <= '1';
					if (initstate = init_get_sic) then
						controllerstate <= no_card;
					else
						clock_enable <= '0';
					end if;
				
				when others =>
					controllerstate <= no_card;
					
			end case;
		end if;
	end process;
	
end architecture rtl; -- of sdcard_controller
